package RFC::Multiplex;
#$Id: Multiplex.pm,v 1.69 2015/05/18 08:11:30 sqko Exp $

###################################################################
#
#
#     RFC::Multiplex
# 
#  This class defines the control of multiple gasses to the same
#  MFC. To obtain a Multiplexer instance call the constructor:
#
#  $id = RFC::Multiplex->new($name,$rig)
#
#  The constructor can also be called with an additional argument
#  which must then be a RFC::MFC (or derived) device instance:
#
#  $mfc = RFC::MFC->new($name,$rig);
#
#  $id = RFC::Multiplex->new($name,$rig,$mfc);
#
#  
#  All Multiplex objects have the following member functions (excluding 
#  those derived from RFC::BaseDevce and Debug.pm):
#
#  $id->attatch_controler($)   Attatches an RFC::MFC instance to the
#                              multiplexer (if not included in the
#                              call of the constructor).
#  $id->set($)                 Sets the multiplexer to select the
#                              specified gas name.
#  $id->channel($)             Returns the relay channel for the 
#                              specified gas name.
#  $id->gasses()               Returns a list of possible gasses.
#  $id->gas()                  Returns the name of the selected gas (NB NOT device name).
#                              If supplied with an optional argument this is used
#                              as a data structure to access if the device had
#                              already been read (similar to the read() and readstring())
#                              member functions inherited rom BaseDevice.
#  $id->gas_name()             Returns the device name of the selected gas.
#                              If supplied with an optional argument this is used
#                              as a data structure to access if the device had
#                              already been read (similar to the read() and readstring())
#                              member functions inherited rom BaseDevice.
#  $id->channel($)             Returns the relay channel for the specified
#                              gas name.
#
# Prior to version 4.2.5, the implementation of the RFC::Multiplex class assumed 
# that all gasses were controled from the same relay board (usually a ICP-con relay module)
# Thus all gas relays shared the same device tty and address, but differed
# only in the channel number. As of version 5.0 this is deprecated and will only
# be avaliable for configuration if legacy mode is enabled (it will still work for
# already configured systems).
#
# As of version 4.2.5 and onwards, it is possible to use arbitrary relay
# devices to control a gas multiplexer. To do so, simply configure the relays
# as normal RFC::Relay instances and select the proper names (defined in the
# '$gasname'_device_name key in the configuration section ($gasname being the name of
# the gas line in question).
#
#
###################################################################

use strict;
use warnings;
use RFC::Header;
use SemaforeFile;
use RFC::BaseDevice;
use RFC::Cache;
our @ISA = qw(RFC::BaseDevice);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    $self->{'_controler'} = undef;
    $self->{'_controler'} = shift if (@_);
    $self->{'_type'} = 'Multiplexer';
    $self->{'_setup_tags'} = ['mode','title','show_plot','use_cache'];
    $self->{'_legacy_tags'} = ['mode','tty','address','control_type','title','show_plot','use_cache'];
    $self->{'_desc'} = {
        'mode' => 'Device control mode',
        'tty' => 'Device communication device (ex. ttyS0)',
        'address' => 'Device address',
	'control_type' => 'Device control type',
	'title' => 'Optional title',
        'show_plot'=>'Determines if the current device data is to be shown in the daily data plots',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    $self->{'_default'} = {
        'mode' => 'Automatic',
        'tty' => 'ttyS0',
        'address' => 0,
	'control_type' => 'ICP',
        'show_plot'=>'Yes',
        'use_cache'=>'No',
    };
    $self->{'_values_control_type'} = [(RFC::Device::list_RELAY())];
    $self->{'_section'} = 'gas_multiplexer_'.$self->name;
    bless $self,$class;
    $self->load_conf();
    $self->{'_file'} = SemaforeFile->new($userdir.'rig'.$self->rig->name.'/multiplex_'.$self->name,'/home/celltest/Gas_rig'.$self->rig->name.'.lock');
    unless ($self->{'_file'}->exist)
    {
        ### Create file with rw for all           ###
        ### Yes I know 666 is the number of evil! ###
        $self->{'_file'}->lock_ex;
        $self->{'_file'}->writeline('');
        chmod 0666, $self->{'_file'}->filename;
        $self->{'_file'}->unlock;
    }
    return $self->init;
}

sub DESTROY
{
    my $self = shift;
    $self->{'_file'}->DESTROY() if ($self->{'_file'});
    return undef;
}


sub attatch_controler($)
{
    my $self = shift;
    $self->{'_controler'} = shift;
    return $self->init;
}

sub init
{
    my $self = shift;
    $self->{'_gasses'} = [];
    $self->{'_channels'} = {};
    $self->{'_gasses'} = [$self->{'_controler'}->gasses] if ($self->{'_controler'});
    if ($self->mode =~ /auto/i)
    {
        my $type = $self->get_conf('control_type');
        $type = 'ICP' unless $type;
	foreach my $gas (@{$self->{'_gasses'}})
	{
	    my $dev = $self->rig->get_device('relay',$self->get_conf($gas.'_device_name'));
	    unless ($dev)
	    {
		$dev = $self->rig->init_device('relay',$self->get_conf($gas.'_device_name'));
		$dev->readstring_ignore(1) if ($dev);
	    }
	    unless ($dev)
	    {
		$dev = RFC::Device::new_RELAY($self->rig,'multiplex_'.$self->name.'_'.$gas,$type,$self->{'_tty'},$self->{'_address'},$self->channel($gas));
		$dev->readstring_ignore(1) if ($dev);
	    }
	    $self->die("Eerror: Could not initialise relay device for gas $gas") unless ($dev);
	    $self->{'_channels'}{$gas} = $dev;
	}
    }
    else
    {
	foreach my $gas (@{$self->{'_gasses'}})
	{
	    my $dev = RFC::Device::new_RELAY($self->rig,'multiplex_'.$self->name.'_'.$gas,'Manual');
	    $dev->readstring_ignore(1);
	    $self->{'_channels'}{$gas} = $dev;
	}
    }
    return $self;
}

sub test()
{
    my $self = shift;
    my $ok = 1;
    foreach my $g (@{$self->{'_gasses'}})
    {
	$ok = 0 unless ($self->{'_channels'}{$g}->test);
    }
    return $ok;
}

sub test_string()
{
    my $self = shift;
    my $string = 'Multiplexer '.$self->name.': ';
    my @arr = ();
    my $data = RFC::Cache->new();
    foreach my $g (@{$self->{'_gasses'}})
    {
	push @arr,$g.' => '.$self->{'_channels'}{$g}->read($data);
    }
    $string .= join(', ',@arr)."\n";
    return $string;
}


sub monitor()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    my @arr = ();
    foreach my $g (@{$self->{'_gasses'}})
    {
	push @arr,$g.' => '.$self->{'_channels'}{$g}->read($data);
    }
    return join(', ',@arr)."\n";
}

sub gasses()
{
    my $self = shift;
    $self->init() if (!@{$self->{'_gasses'}} && $self->{'_controler'});
    return @{$self->{'_gasses'}};
}

sub gas_all()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my @gas = ();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $status = $self->{'_channels'}{$g}->read($data);
	next unless ($status == 1);  ## if read returns -1 it is an error value ##
	my $gdev = $self->rig->get_device('gas',$g);
	push @gas, $gdev->gas();
    }
    unless (@gas == 1)
    {
	$self->writedebug("Retrying as no gas appears to be selected!");
	@gas = ();
	## Retry if communication failed ##
	my $d2 = RFC::Cache->new();
	foreach my $g (@{$self->{'_gasses'}})
	{
	    my $status = $self->{'_channels'}{$g}->read($d2);
	    next unless ($status == 1);   ## if read returns -1 it is an error value ##
	    my $gdev = $self->rig->get_device('gas',$g);
	    push @gas, $gdev->gas();
	}
    }
    unless (@gas)
    {
	my $str = 'Warning: No gas was engaged for multiplexer '.$self->name.' for rig '.$self->rig->name;
        $self->rig->warning_mail('Multiplexer_'.$self->name,$str);
	$self->writedebug($str,1);
	return undef;
    }
    if (@gas > 1)
    {
	my $str = 'Alert! More than one gas is selected for multiplexer '.$self->name.' for rig ' . $self->rig->name . '. Selected gasses: '.join(', ',@gas);
        $self->rig->warning_mail('Multiplexer_'.$self->name,$str);
	$self->writedebug($str,1);
	return undef;
    }
    return @gas;
}

sub gas()
{
    my $self = shift;
    my @list = $self->gas_all(@_);
    return $list[0];
}

sub gas_all_name()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my @gas = ();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $status = $self->{'_channels'}{$g}->read($data);
	next unless ($status == 1);  ## if read returns -1 it is an error value ##
	my $gdev = $self->rig->get_device('gas',$g);
	push @gas, $gdev->name();
    }
    unless (@gas == 1)
    {
	$self->writedebug("Retrying as no gas appears to be selected!");
	@gas = ();
	## Retry if communication failed ##
	my $d2 = RFC::Cache->new();
	foreach my $g (@{$self->{'_gasses'}})
	{
	    my $status = $self->{'_channels'}{$g}->read($d2);
	    next unless ($status == 1);   ## if read returns -1 it is an error value ##
	    my $gdev = $self->rig->get_device('gas',$g);
	    push @gas, $gdev->name();
	}
    }
    unless (@gas)
    {
	my $str = 'Warning: No gas was engaged for multiplexer '.$self->name.' for rig '.$self->rig->name;
        $self->rig->warning_mail('Multiplexer_'.$self->name,$str);
	$self->writedebug($str,1);
	return undef;
    }
    if (@gas > 1)
    {
	my $str = 'Alert! More than one gas is selected for multiplexer '.$self->name.' for rig ' . $self->rig->name . '. Selected gasses: '.join(', ',@gas);
        $self->rig->warning_mail('Multiplexer_'.$self->name,$str);
	$self->writedebug($str,1);
	return undef;
    }
    return @gas;
}

sub gas_name()
{
    my $self = shift;
    my @list = $self->gas_all_name(@_);
    return $list[0];
}

sub readstring()
{
    my $self = shift;
    return undef if ($self->readstring_ignore);
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my $res = undef;
    my @t = $self->get_time();
    my $t = join(':',@t);
    my @gas = $self->gas_all_name($data);
    my $gas = join(',',@gas);
    $gas = 'NONE' unless (@gas);
    return join(' ',$self->name.'_selected_gas',$t,time,$gas);
}

sub read_nocache
{
    my $self = shift;
    my $gas = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    unless ($gas)
    {
	return $self->gas($data);
    }
    unless (exists($self->{'_channels'}{$gas}))
    {
	$self->writedebug("Warning, None of the gasses for multiplexer ".$self->name." was found matching gas $gas for rig ".$self->rig->name);
	return undef;
    }
    my $res = $self->{'_channels'}{$gas}->read($data);
    $self->writedebug("Status for multiplexer ".$self->name." for gas $gas on channel ".$self->channel($gas)." was $res") if ($self->debug > 2);
    $res = 0 unless (defined($res));
    my $str = time .' '.$self->name. ' '.$res;
    $self->rig->kanaldata($str) unless ($self->readstring_ignore);
    $self->set_cache($res);
    return $res;
}

sub set()
{
    my $self = shift;
    my $gas = shift;
    $self->writedebug("Setting multiplexer ".$self->name." to $gas") if ($self->debug > 2);
    unless (exists($self->{'_channels'}{$gas}))
    {
	$self->die("Error: None of the gasses for multiplexer ".$self->name." was found matching gas $gas for rig ".$self->rig->name);
	return undef;
    }
    my $data = RFC::Cache->new();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $status = $self->{'_channels'}{$g}->read($data);
	## Only set relays to 0 for engaged relays (saves serial bus bandwidth) ##
	next if ($status == 0 || $g eq $gas);
	$self->{'_channels'}{$g}->set(0,@_);
    }
    $self->{'_channels'}{$gas}->set(1,@_);
    $self->rig->change_config_value('gas_'.$gas,'mode','Automatic');
    ### Set gas flows for not engaged gasses to 0 and mark as manual ###
    $self->writedebug("Setting non engaged gas flows to 0") if ($self->debug > 2);
    foreach my $g (@{$self->{'_gasses'}})
    {
	next if ($g eq $gas);
	$self->rig->change_config_value('gas_'.$g,'mode','Manual');
	my $dev = $self->rig->get_device('gas',$g);
	if ($dev)
	{
	    $dev->setflow(0,@_);
	}
	else
	{
	    $self->rig->proglog("Error: could not get device for gas $gas when changing multiplexer ".$self->name);
	}
    }
    ### Set gas for controler ###
    unless ($self->{'_controler'})
    {
	foreach my $d ($self->rig->list_devices('MFC'))
	{
	    my $dev = $self->rig->get_device('MFC',$d);
	    $self->{'_controler'} = $d if ($dev && $dev->multiplexer == $self->name);
	}
    }
    $self->rig->change_config_value('controler_'.$self->{'_controler'}->name,'gas',$gas) if ($self->{'_controler'});
    ### Reset error lock ###
    $self->rig->unset_warning('Multiplexer_'.$self->name);
    $self->notify('set',$gas);
}

sub print_config()
{
    my $self = shift;
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Mode: '.$self->mode;
    push @arr,'Type: '.$self->type;
    push @arr,'TTY: '.$self->{'_tty'};
    push @arr,'Address: '.$self->{'_address'};
    my @chs = ();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $dev = $self->rig->init_device('relay',$self->get_conf($g.'_device_name'));
	if ($dev)
	{
	    push @chs,$g.'->'.$dev->print_config;
	}
	else
	{
	    push @chs,$g.'->'.$self->channel($g);
	}
    }    
    push @arr,'channels: '.join(',',@chs);
    return join (', ',@arr);
}

sub debug()
{
    my $self = shift;
    $self->SUPER::debug(@_);
    if (@_)
    {
        #### Pass on debug level to all objects honoring this ###
	if ($self->mode =~ /auto/i)
	{
	    foreach my $gas (@{$self->{'_gasses'}})
	    {
		$self->{'_channels'}{$gas}->debug($self->{'_debug'}) if ($self->{'_channels'}{$gas});
	    }
	}
    }
    return $self->{'_debug'};
}

sub channel($)
{
    my $self = shift;
    my $gas = shift;
    my $ch = $self->get_conf($gas);
    return $ch;
}

