package RFC::Typecast;
#$Id: Typecast.pm,v 1.9 2015/05/11 09:12:52 sqko Exp $
#################################################################
#
#
#       RFC::Typecast
#
#
# The RFC::Typecast is a special filter device. 
# It can convert one underlying device to an other type.
# 
# Typcast devices should not normally be nescesarry, but in specific 
# cases it may be nescesarry to do a typecast.
#
# Typecast devices have a simplified interface as they only implement
# the BaseDevice classes.
#
# An exception to this is the notify function which allows the conversion
# of the originating command to the one specified by the 'notify_function'
# value. Default is the 'set' command.
#
# The possible values for this variable is determined by inspecting any
# devices attatched to the Typecast device. If (one or more) attatched 
# devices are found, the first device is asked for potential notify 
# functions by calling the list_set_functions() on it.
#
# As opposed to most other devices, the readstring function simply returns
# the empty string as the value of the typecast device will be equal to the
# unederlying device.
#
#################################################################
use strict;
use warnings;
use RFC::Filter;
use RFC::Header;
use RFC::Cache;
our @ISA = qw(RFC::Filter);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}


sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    $self->{'_type'} = 'Typecast_device';
    $self->{'_setup_tags'} = ['class','mode','device_type','device','output_type','output_device','title','show_plot','notify_function'];
    $self->{'_desc'} = {
        'class' => 'Device class',
        'mode' => 'Device control mode',
        'device_type' => 'Type of raw device',
        'device' => 'Raw device',
        'title' => 'Optional title',
        'output_type' => 'Type of device to convert to',
        'output_device' => 'Optional output device',
        'show_plot'=>'Determines if the current device data is to be shown in the daily data plots',
	'notify_function' => 'Function to call on the output device and any listening devices if a notify event is called on current device',
    };
    $self->{'_default'} = {
        'class' => 'Typecast',
        'device_type' => 'simplechannel',
        'output_type' => 'gas',
        'device' => '',
        'output_device' => '',
        'spline' => '',
        'read_function' => 'read',
        'show_plot'=>'Yes',
	'notify_function' => 'set',
    };
    bless $self,$class;
    $self->readstring_ignore(1);
    foreach my $tag ($self->setup_tags)
    {
        $self->{'_'.$tag} = $self->get_default($tag);
        my $conf = $self->get_conf($tag);
        $self->{'_'.$tag} = $conf if (defined($conf) && $conf =~ /\w/);
    }
    $self->{'_read_names'} = ['read'];
    $self->get_set_function_names();
    $self->{'_values_output_type'} = [$self->rig->list_types_nofilter()];
    my @list = ();
    foreach ($self->rig->avaliable_devices($self->{'_output_type'}))
    {
        push @list,$_ unless ($_ eq $self->name()); ## Do not include self ##
    }
    $self->{'_values_output_device'} = ['',@list];
    $self->rig->register_device($self->{'_output_type'},$self);
    ### Note, do NOT run init as part of constructor! ###
    return $self;
}

sub setup_tags()
{
    ###############################################
    ### setup_tags()                            ###
    ### Returns a list of acceptable setup tags ###
    ### Can be used by a application to easialy ###
    ### get the setup tags and values           ###
    ###############################################
    my $self = shift;
    my @res = ();
    @res = @{$self->{'_setup_tags'}} if ($self->{'_setup_tags'});
    return (@res,'comments');
}

sub get_set_function_names()
{
    my $self = shift;
    if ($self->{'_odev'})
    {
	$self->{'_values_notify_function'} = [$self->{'_odev'}->list_set_functions()];
	return;
    }
    my @list = $self->list_observers();
    if (@list)
    {
	$self->{'_values_notify_function'} = [${$list[0]}->list_set_functions()];
    }
    return $self;
}

sub read
{
    my $self = shift;
    return $self->{'_dev'}->read(@_);
}

sub print_config()
{
    my $self = shift;
    $self->init();
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Type: '.$self->type;
    push @arr,'Input device type: '.$self->{'_device_type'};
    push @arr,$self->{'_dev'}->print_config();
    push @arr,'Output device type: '.$self->{'_output_type'};
    push @arr,$self->{'_odev'}->print_config() if ($self->{'_odev'});
    return join(', ',@arr);
}

sub test_string
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
	$data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug) if ($self->debug);
    }
    $self->check_cache($data);
    my @arr = ();
    $self->init();
    my $string = 'Typecast '.$self->name.': ';
    my $val = $self->read($data);
    push @arr,'Value: '.$val;
    push @arr,'Converts from '.$self->{'_device_type'}.' to '.$self->{'_output_type'};
    push @arr,'Forwarding to '.$self->{'_notify_function'};
    $string .= join(', ',@arr)."\n";
    return $string;
}


sub local_init()
{
    my $self = shift;
    if (!$self->{'_odev'} && $self->{'_output_type'} && $self->{'_output_device'})
    {
        ### get device if it already exist                                          ###
        ### It may have been created after SPDEV constructor ran and call of init() ###
        my $dev = $self->rig->get_device($self->{'_output_type'},$self->{'_output_device'});
        ### Initialise it if it doesn't ###
        unless ($dev)
        {
            $dev = $self->rig->init_device($self->{'_output_type'},$self->{'_output_device'});
            $dev->readstring_ignore(1) if ($dev);
        }
        $self->{'_odev'} = $dev if ($dev);
        $dev->debug($self->debug);
	$dev->set_readonly();
        $self->die("Could not initialise device of type '".$self->{'_output_type'}."' with name: '".$self->{'_output_device'}."'") unless ($self->{'_odev'});
    }
    my @arr = ('');
    if ($self->{'_dev'})
    {
	@arr = $self->{'_dev'}->list_set_functions();
	$self->{'_dev'}->attatch_observer(\$self);
    }
    $self->{'_set_fn_names'} = \@arr;
    $self->get_set_function_names();
    return $self;
}

sub attatch_observer($)
{
    my $self = shift;
    return &RFC::BaseDevice::attatch_observer($self,@_);
}


sub detatch_observer($)
{
    my $self = shift;
    return &RFC::BaseDevice::detatch_observer($self,@_);
}

sub notify($)
{
    my $self = shift;
    return &RFC::BaseDevice::notify($self,@_);
}

sub observe
{
    my $self = shift;
    $self->init();
    my $cmd = shift;
    ## Exchange notify command with specified function name ##
    $cmd = $self->{'_notify_function'};
    my $res = $self->notify($cmd,@_);
    return $res unless ($self->{'_odev'});
    return $self->{'_odev'}->$cmd(@_);
}

